package com.perforce.maven.scm.provider.p4.util;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.List;

import org.apache.maven.scm.ScmException;
import org.apache.maven.scm.ScmFileSet;
import org.apache.maven.scm.ScmTag;
import org.apache.maven.scm.ScmVersion;
import org.codehaus.plexus.util.ReaderFactory;
import org.codehaus.plexus.util.StringUtils;
import org.codehaus.plexus.util.xml.pull.XmlPullParserException;

import com.perforce.maven.scm.provider.p4.repository.P4ScmProviderRepository;
import com.perforce.maven.scm.provider.p4.settings.Settings;
import com.perforce.maven.scm.provider.p4.settings.io.xpp3.P4MavenXpp3Reader;
import com.perforce.p4java.client.IClient;
import com.perforce.p4java.core.file.FileSpecBuilder;
import com.perforce.p4java.core.file.FileSpecOpStatus;
import com.perforce.p4java.core.file.IFileSpec;
import com.perforce.p4java.exception.P4JavaException;

public class P4Utils
{
    private P4Utils()
    {
    }

    /**
     * Gets the canonical repo path.
     *
     * @param repoPath the repo path
     * @return the canonical repo path
     */
    public static String getCanonicalRepoPath( String repoPath )
    {
        if ( repoPath == null )
        {
            return null;
        }

        if ( repoPath.endsWith( "/" ) )
        {
            repoPath += "...";
        }
        else if ( !repoPath.endsWith( "/..." ) )
        {
            repoPath += "/...";
        }

        if ( repoPath.contains( " " ) )
        {
            repoPath = "\"" + repoPath + "\"";
        }

        return repoPath;

    }

    /**
     * Perforce wildcards expansion.
     *
     * @param filePath the file path
     * @return the string
     */
    public static String encodeWildcards( String filePath )
    {
        String path = new String();
        if ( filePath != null )
        {
            path =
                filePath.replaceAll( "%", "%25" ).replaceAll( "\\*", "%2A" ).replaceAll( "#", "%23" ).replaceAll( "@",
                                                                                                                  "%40" );
        }
        return path;
    }

    /**
     * From a given local file, looks for its remote path
     *
     * @param localFile localPath
     * @return remote path
     */
    public static String getRepoLocation( IClient client, File localFile )
        throws P4JavaException
    {

        List<IFileSpec> fileSpecs =
            client.where( FileSpecBuilder.makeFileSpecList( encodeWildcards( localFile.getAbsolutePath() ) ) );

        for ( IFileSpec fileSpec : fileSpecs )
        {
            if ( !StringUtils.isBlank( fileSpec.getDepotPathString() ) )
            {
                return fileSpec.getDepotPathString();
            }
        }

        return null;
    }

    public static boolean isRepoPathValid( IClient client, String repoPath )
        throws P4JavaException
    {
        List<IFileSpec> fileSpecs = client.where( FileSpecBuilder.makeFileSpecList( getCanonicalRepoPath( repoPath ) ) );

        if ( fileSpecs.size() != 0 )
        {
            // check for error
            IFileSpec fileSpec = fileSpecs.get( 0 );
            if ( fileSpec.getOpStatus() == FileSpecOpStatus.VALID )
            {
                return true;
            }
        }

        return false;
    }

    public static String scmVersion2SuffixString( ScmVersion version )
    {
        String versionSuffix = "";
        if ( version != null && !StringUtils.isBlank( version.getName() ) )
        {
            versionSuffix = "@" + version.getName();
        }

        return versionSuffix;
    }

    // //////////////////////////////////////////////////////////////////////////////////////////////////////////
    // ScmFileSet to IFileSpec conversion
    // //////////////////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * Convert ScmFileSet to List<IFileSpec> for
     *
     * @param fileSet
     * @return
     * @throws ScmException
     */
    public static List<IFileSpec> scmFileSet2P4FileSpecsWithNop4WildCard( ScmFileSet fileSet, ScmVersion version )
        throws ScmException
    {

        String versionSuffix = scmVersion2SuffixString( version );

        List<IFileSpec> fileSpecs = new ArrayList<IFileSpec>();
        try
        {
            List<File> fileList = fileSet.getFileList();
            if ( fileList != null )
            {
                HashSet<String> filePaths = new HashSet<String>();
                for ( File file : fileList )
                {
                    if ( ".".equals( file.getPath() ) )
                    {
                        // ignore child path "." since it is the same as parent dir
                        // this happens when issue mvn scm:command from command line
                        continue;
                    }

                    if ( !file.isAbsolute() )
                    {
                        // It is assumed that ScmFileSet does validate its fileList which must be relative to its
                        // basedir
                        file = new File( fileSet.getBasedir(), file.getPath() );
                    }
                    filePaths.add( P4Utils.encodeWildcards( file.getCanonicalPath() + versionSuffix ) );

                }
                if ( !filePaths.isEmpty() )
                {
                    fileSpecs = FileSpecBuilder.makeFileSpecList( filePaths.toArray( new String[filePaths.size()] ) );
                }
            }
        }
        catch ( IOException e )
        {
            throw new ScmException( e.getLocalizedMessage(), e );
        }

        return fileSpecs;

    }

    public static List<IFileSpec> scmFileSet2P4FileSpecsWithNoP4WildCard( ScmFileSet fileSet )
        throws ScmException
    {
        return scmFileSet2P4FileSpecsWithNop4WildCard( fileSet, null );
    }

    public static List<IFileSpec> scmFileSet2P4FileSpecs( ScmFileSet fileSet, ScmVersion version )
        throws ScmException
    {
        List<IFileSpec> fileSpecs = scmFileSet2P4FileSpecsWithNop4WildCard( fileSet, version );

        if ( fileSpecs.isEmpty() )
        {
            if ( StringUtils.isBlank( fileSet.getIncludes() ) && StringUtils.isBlank( fileSet.getExcludes() ) )
            {
                String versionSuffix = scmVersion2SuffixString( version );
                if ( StringUtils.isBlank( versionSuffix ) )
                {
                    versionSuffix = "/...";
                }
                String filePath = fileSet.getBasedir().getAbsolutePath() + versionSuffix;

                // very hacky
                if ( version instanceof ScmTag )
                {
                    filePath = "@" + version.getName();
                }
                fileSpecs = FileSpecBuilder.makeFileSpecList( new String[] { filePath } );
            }
        }

        return fileSpecs;

    }

    public static List<IFileSpec> scmFileSet2P4FileSpecs( ScmFileSet fileSet )
        throws ScmException
    {
        return scmFileSet2P4FileSpecs( fileSet, null );
    }

    // //////////////////////////////////////////////////////////////////////////////////////////////////////////
    // P4Maven Settings
    // //////////////////////////////////////////////////////////////////////////////////////////////////////////

    private static final String P4MAVEN_SETTINGS_FILENAME = "p4maven-settings.xml";

    public static final File DEFAULT_SETTINGS_DIRECTORY = new File( System.getProperty( "user.home" ), ".scm" );

    private static File settingsDirectory = DEFAULT_SETTINGS_DIRECTORY;

    private static Settings settings;

    /**
     * For unittest only
     */
    public static synchronized void resetSettings()
    {
        settings = null;
    }

    public static void setP4MavenSettingsDirecotry( File location )
    {
        settingsDirectory = location;
    }

    public static synchronized Settings getSettings()
    {
        if ( settings == null )
        {
            settings = readSettings();
        }
        return settings;
    }

    private static Settings readSettings()
    {
        File settingsFile = getSettingsFile();

        Settings settings = new Settings();

        if ( settingsFile.exists() )
        {
            P4MavenXpp3Reader reader = new P4MavenXpp3Reader();
            try
            {
                settings = reader.read( ReaderFactory.newXmlReader( settingsFile ) );
            }
            catch ( FileNotFoundException e )
            {
                // nop
            }
            catch ( IOException e )
            {
                // nop
            }
            catch ( XmlPullParserException e )
            {
                String message = settingsFile.getAbsolutePath() + " isn't well formed. SKIPPED." + e.getMessage();

                System.err.println( message );
            }
        }

        // override if found from system properties
        String jobs = System.getProperty( P4ScmProviderRepository.P4_JOBS_PROPERTY );
        if ( jobs != null )
        {
            settings.setJobs( jobs );
        }

        // nullify value in the p4maven-settings.xml, we dont want user to set it there
        settings.setClientName( null );

        String clientName = System.getProperty( P4ScmProviderRepository.P4_CLIENT_PROPERTY );
        if ( clientName != null )
        {
            settings.setClientName( clientName );
        }

        String charset = System.getProperty( P4ScmProviderRepository.P4_CHARSET_PROPERTY );
        if ( charset != null )
        {
            settings.setCharset( charset );
        }

        //
        // String p4PortEnv = System.getenv( P4ScmProviderRepository.P4_PORT_PROPERTY );
        // if ( StringUtils.isBlank( settings.getP4Port() ) && !StringUtils.isBlank( p4PortEnv ) )
        // {
        // settings.setP4Port( p4PortEnv );
        // }
        //
        // String p4PortProp = System.getProperty( P4ScmProviderRepository.P4_PORT_PROPERTY );
        // if ( !StringUtils.isBlank( p4PortProp ) )
        // {
        // settings.setP4Port( p4PortProp );
        // }

        return settings;
    }

    public static List<String> getJobs()
    {
        String jobs = getSettings().getJobs();
        if ( jobs == null )
        {
            jobs = "";
        }

        String[] tokens = StringUtils.split( jobs );

        return new ArrayList<String>( Arrays.asList( tokens ) );
    }

    private static File getSettingsFile()
    {
        return new File( settingsDirectory, P4MAVEN_SETTINGS_FILENAME );
    }

    //for unit test only
    public static void setSettingsDirectory( File directory )
    {
        settingsDirectory = directory;
    }

}
